/*
 * The codes are the simulations for fBm queues from the paper:
 * Performance evaluation and service rate provisioning for a queue with fractional Brownian input.
 * The Main function is getProb(), which returns the overflow probability.
 */
 
#include <stdio.h>
#include <math.h>
#include "hosking.h"

#define OBS 11		/*observation number*/
#define SD 2.23		/*value of t_(0.025,10)*/

double calConf(double *upper, double *lower, double prob[]){
	/*		upper:	upper bound of the confidence interval      */
	/*		lower:	upper bound of the confidence interval		*/
	/*		prob[]:	array of the input values					*/
	/*		OBS:	no. of the observations (length og prob[])	*/
	/*		SD:		degree of freedom as t_{0.025,OBS-1}		*/
	int i;
	double eve, secondM, conf;
	//eve = mean; secondM = second moment; conf = confidence interval
	eve = 0.0;
	for(i=0; i<OBS; i++){
		eve += prob[i];
	}
	eve /= OBS;
	secondM = 0.0;
	for(i=0; i<OBS; i++){
		secondM += (prob[i]-eve) * (prob[i]-eve);
	}
	secondM /= (OBS-1);
	conf = SD*sqrt(secondM/OBS);
	*upper = eve + conf;
	*lower = eve - conf;
	return eve;		//return the mean of the input array
}

double calConf(double *error, double prob[]){
	/*		error:	the confidence interval						*/
	/*		prob[]:	array of the input values					*/
	/*		OBS:	no. of the observations (length og prob[])	*/
	/*		SD:		degree of freedom as t_{0.025,OBS-1}		*/
	int i;
	double eve, secondM;
	//eve = mean; secondM = second moment
	eve = 0.0;
	for(i=0; i<OBS; i++){
		eve += prob[i];
	}
	eve /= OBS;
	secondM = 0.0;
	for(i=0; i<OBS; i++){
		secondM += (prob[i]-eve) * (prob[i]-eve);
	}
	secondM /= (OBS-1);
	*error = SD*sqrt(secondM/OBS);
	return eve;		//return the mean of the input array
}


void getProb(int n, double H,double mu, FILE *fc, FILE *ft, char* inF, int xN, double iniX, double v, double sigmaS, double deltaT){
	/*		function that gets the probability for given datafile       */
	/*		n		determines the sample size N as N=2^(*n)			*/
	/*		mu		net mean input										*/
	/*		*fc		output file with extension csv with type FILE		*/ 
	/*		*ft		output file with extension txt with type FILE		*/ 
	/*		*inF	name of the input file containing fGn sequences		*/
	/*		xN		size of buffer threshold x							*/
	/*		iniX	initial value/scale of x							*/
	/*		v		variance in one unit time							*/
	/*		sigmaS	the variance in one second							*/
	/*		deltaT	one unit of time representing deltaT second			*/
	/*		OBS		number of observations desired						*/
	int i, j, m, bound, **prCount,intN;
	double fgn, prob, error, que, **pr, *x, nomUnit, mW;
	FILE *in;

	in = fopen(inF,"r");	
	bound = pow(2.0,n);		
	intN = floor(pow(2.0,n)/(OBS+1.0));
	nomUnit = sqrt(sigmaS/v)*pow(deltaT,H);	
	mW = mu*deltaT/nomUnit;

	prCount = (int **) calloc(xN, sizeof(int*));
	pr = (double **) calloc(xN, sizeof(double*));
	for(i=0;i<xN;i++){
		prCount[i] = (int *) calloc(OBS, sizeof(int));
		pr[i] = (double *) calloc(OBS, sizeof(double));
	}
	x = (double *) calloc(xN, sizeof(double));

	/* form the array of x values */
	x[0] = iniX;
	for(i=1; i < xN; i++){
		x[i] = x[i-1] +iniX;
	}

	for(j=0; j<xN; j++){
		for(i=0; i<OBS; i++){
			prCount[j][i] = 0;
		}
	}

	/* start with an empty queue and do intN intervals without obseveration */
	que = 0.0;
	for(i=0; i < intN;i++){
		fscanf(in,"%lf",&fgn);
		que = que + fgn + mW;
		if(que<0.0)
			que = 0.0;
	}

	/* start to observe the queue with around intN*OBS intervals */
	m = 0;	// index of the observation
	for(i=intN; i < bound; i++){
		fscanf(in,"%lf",&fgn);
		que = que + fgn + mW;
		/* eunsure that queue size is not negative */
		if(que<0.0)
			que = 0.0;
		/* start a new observation */	
		if(i%intN==0 && i!=intN)
			m++;
		for(j=0; j < xN; j++){
			/* queue size is larger than the threshold */
			if(que>(x[j]/nomUnit))
				prCount[j][m]++;
		}			
	}

	
	for(j=0; j<xN; j++){
		/* obatin the overflow probability for each observation */
		for(i=0;i<OBS;i++){	
			pr[j][i] = (double)prCount[j][i]/(double)intN;
		}
	}

	for(j=0;j<xN;j++){
		/* obtain the confidence intervals */
		prob = calConf(&error, pr[j]);

		/* write the output files */
		fprintf(fc, "%lf,%lf,%lf,%lf\n", mu, x[j], prob, error);
		for(i=0;i<OBS;i++)
			fprintf(ft, "%lf,%lf,%e\n", mu, x[j], pr[j][i]);
	}
	fclose(in);
}

//This is an example of simulating fBm queue.
void main(){
	FILE *fc, *ft;
	/* path of the file containing the fGn sequence */
	char *in = "Sim data/data/n 22 v 1 H 0-8.txt";

	/* path of the output files: fc (csv file), ft (txt file) */
	fc = fopen("H08 s1 d0_01.csv","w");
	ft = fopen("H08 s1 d0_01 detail.csv","w");

	/* one example of obtaining the overflow probability */
	getProb(22,0.8,-0.5,fc,ft,in,1000,0.1,1,1,0.01);
	fclose(fc);
	fclose(ft);
}